import scala.io.Source
import scala.sys.process._
import java.lang.management.ManagementFactory
import java.io.File

object mfn {
  var nmzdCASP: Seq[String] = Seq.empty

  /* library */
  val casp2sugar = "lib/casp2sugar.pro"
  val sugar2aspPro = "lib/trans_asp.pro"
  val oelp = "lib/oe.lp"
  val swipl = "lib/swipl.sh"

  /* files */
  val pid = ManagementFactory.getRuntimeMXBean.getName.split("@").head
  var prefix = s"m${pid}-"
  var prmlConstFile = prefix + "Step2-csp.lp"
  var pureASPFile = prefix + "Step2-asp.lp"
  var sugarConstFile = prefix + "Sugar-before.csp"
  var nmzdSugarConstFile = prefix + "Sugar-after.csp"
  var constASPFile = prefix + "Step3-csp.lp"
  var gndASPFile = prefix + "Step4-asp.gnd"
  var aspFiles: Seq[String] = Seq.empty
  val allTmpFiles = Seq(prmlConstFile,pureASPFile,sugarConstFile,nmzdSugarConstFile,constASPFile,gndASPFile)
  val subTmpFiles = Set(prmlConstFile,pureASPFile,sugarConstFile)

  /* options */
  val claspOp = """-C([a-z,=-]+)""".r
  val numR = """([0-9]+)""".r
  val strR = """([a-z]+)""".r
  val gringoOp = """-G([a-z=-]+)""".r
  var opsClasp: Seq[String] = Seq.empty
  var opsGringo: Seq[String] = Seq.empty
  var opsSugar: Seq[String] = Seq("arity=3")
  var xcsp = false
  var keep = false

  def show(file: String) {
    s"cat ${file}" !
  }

  def runGringo(file: String, ops: String = "") {
    println("=== Step 1 ==============================================")
    println("generate grounded plain taxt format of CASP")
    println(s"gringo ${ops} ${file}")
    nmzdCASP = s"gringo ${ops} ${file}" lines_!

    println()
  }

  def runGringoT(file: String) {
    runGringo(file,"-t")
    val v = """.*variable.*""".r
    val c = """.*collection.*""".r
    val con = """.*constraint.*""".r
    println("=== Step 2 ==============================================")
    println(s"split into (a) ASP Part, (b) Constraint Part")
    println(s"\t(a) ${pureASPFile}\n\t(b) ${prmlConstFile}")
    val out1 = new java.io.PrintWriter(prmlConstFile)
    val out2 = new java.io.PrintWriter(pureASPFile)
    for (line <- nmzdCASP) {
      line match {
        case v() | c() | con() => out1.println(line)
        case n => out2.println(line)
      }
    }
    out1.flush; out1.close
    out2.flush; out2.close
    nmzdCASP = Seq.empty

    println()
  }

  def prml2sugar {
    println("=== Step3-1 ==============================================")
    println(s"(n-ary ASP) ===> (n-ary Sugar)")
    println(s"\t(b) ${prmlConstFile}\n\t(b') ${sugarConstFile}")
    println(s"swipl -c ${prmlConstFile} ${casp2sugar} > ${sugarConstFile}")


    s"swipl -c ${prmlConstFile} ${casp2sugar}" #> new java.io.File(sugarConstFile) !

    println()
  }

  def runSugar(file: String) {
    val pro = if (xcsp) "" else "-prolog"
    val cmd = s"sugar ${pro} -o ${opsSugar.mkString(",")} -translate prolog -output ${nmzdSugarConstFile} ${file}"
    println("=== Step3-2 ==============================================")
    println(s"(n-ary Sugar) ===> (3-ary Sugar)")
    println(s"\t(n-ary Sugar) ${sugarConstFile}\n\t(3-ary Sugar)${nmzdSugarConstFile}")
    println(cmd)

    cmd !

    println()
  }

  def sugar2asp {
    println("=== Step3-3 ==============================================")
    println(s"(3-ary Sugar) ===> (3-ary ASP)")
    println(s"\t(3-ary Sugar) ${nmzdSugarConstFile}\n\t(3-ary ASP) ${constASPFile}")
    println(s"${swipl} ${nmzdSugarConstFile}")

    val lines = s"${swipl} ${nmzdSugarConstFile}" lines_!
    val out = new java.io.PrintWriter(constASPFile)
    for (line <- lines) {
      val converted = line.replace("$","_").replace("\'","").toLowerCase
      out.println(converted)
    }
    out.flush
    out.close

    println()
  }

  def solve {
    val gndCMD = s"gringo ${opsGringo.mkString(" ")} ${aspFiles.mkString(" ")}"
    println("=== Step4 =================================================")
    println(gndCMD)


    gndCMD #> new java.io.File(gndASPFile) !

    println()

    val clspCMD = s"clasp ${opsClasp.mkString(" ")} ${gndASPFile}"
    println("=== Step5 =================================================")
    println(clspCMD)


    clspCMD !
  }
  
  def encode(casp: String) {
    runGringoT(casp)
    prml2sugar
    runSugar(sugarConstFile)
    sugar2asp    
    aspFiles = Seq(pureASPFile,constASPFile,oelp)
  }

  def encodeXCSP(xcspFile: String) {
    runSugar(xcspFile)
    sugar2asp
    aspFiles = Seq(constASPFile,oelp)
  }


  def parseOptions(arguments: List[String]): List[String] = arguments match {
    case "-h" :: rest => {
      parseOptions(rest)
    }
    case claspOp(option) :: numR(num) :: rest => {
      opsClasp = opsClasp :+ s"${option} ${num}"
      println("Clasp Option2: " + s"${option} ${num}")
      parseOptions(rest)
    }
    case claspOp(option) :: strR(str) :: rest => {
      opsClasp = opsClasp :+ s"${option} ${str}"
      println("Clasp Option2: " + s"${option} ${str}")
      parseOptions(rest)
    }
    case claspOp(option) :: rest => {
      opsClasp = opsClasp :+ option
      println("Clasp Option: " + option)
      parseOptions(rest)
    }
    case gringoOp(option) :: rest => {
      opsGringo = opsGringo :+ option
      println("Gringo Option: " + option)
      parseOptions(rest)
    }
    case "-no_simp" :: rest => {
      opsSugar = opsSugar :+ "no_simp"
      parseOptions(rest)
    }
    case "-xcsp" :: rest => {
      xcsp = true
      parseOptions(rest)
    }
    case "-keep" :: rest => {
      keep = true
      parseOptions(rest)
    }


    case _ =>
      arguments
  }

  def removeTmpFiles {
    for (file <- allTmpFiles if (!xcsp || !subTmpFiles.contains(file))) { 
      s"rm ${file}" !
    }
  }

  var inputFile: String = null

  def main(args: Array[String]) {
    println("=======================================================")
    println("This is muffin batch program written in Scala.")
    println(s"PID: ${pid}")
    println

    parseOptions(args.toList) match {
      case file :: Nil => {
        inputFile = file
      }
      case _ => {

        val helpMessage = 
"""=== Help ===
Usage: ./mfn [options] [inputFile]

-no_simp            off Tseitin translation in Sugar
-keep               keep all temporal files
-xcsp               accept xcsp format
-C<clasp options>   any clasp options
-G<gringo options>  any gringo options in Step5

Ex1. enumerate all answer sets without display. with clasp stats. no Tseitin translation. 
$ ./mfn -C-n 0 -C--quiet -C-s -no_simp examples/alldiff.lp

Ex2. solve xcsp with "jumpy". keep temporal files. print gringo stats.
$ ./mfn -xcsp -G--gstats -C--configuration=jumpy -keep examples/queens-12.xml"""
        println(helpMessage)
        System.exit(1)
      }
    }
    if (!xcsp) encode(inputFile) else encodeXCSP(inputFile)
    solve

    if (!keep) removeTmpFiles
  }

}
